from unittest import TestCase

from zabbixasync.sender import AsyncSender, ItemData


class TestPacket(TestCase):
    def test_packet_creation(self):
        expected = b'ZBXD\x01[\x00\x00\x00\x00\x00\x00\x00\
{"request": "sender data", "data": \
[{"host": "localhost", "key": "test", "value": 123456}]}'

        data = ItemData('localhost', 'test', 123456)
        items = [data]
        sender = AsyncSender('server')
        packet = sender._create_packet(items)

        self.assertEqual(packet, expected)



from dataclasses import dataclass
import json
import logging
import re
import struct
from pathlib import Path
from typing import Any, Optional, Sequence, Tuple
import asyncio
import ssl
import zlib


logger = logging.getLogger(__name__)
ZABBIX_RETURN_REGEX = re.compile(
    r'processed: (\d+); failed: (\d+); total: (\d+); seconds spent: (\d+\.\d+)'
    )


class ItemData(dict):
    """
    Dictionary representing a trapper item.
    """
    __slots__ = ['__dict__']

    # implementing a dictionary helps to json.dumps
    def __init__(self,
                 host: str,
                 key: str,
                 value: Any,
                 clock: Optional[int] = None,
                 ns: Optional[int] = None):
        dict.__init__(self, host=host, key=key, value=value)
        if clock is not None:
            self['clock'] = clock
        if ns is not None:
            self['ns'] = ns


@dataclass
class ZabbixResponse:
    processed: int
    failed: int
    total: int
    seconds_spent: float
    response: str


class AsyncSender():
    __slots__ = [
        "server",
        "port",
        "timeout",
        "tls_connect",
        "tls_psk_identity",
        "tls_psk",
        "tls_psk_file",
        "tls_server_name",
    ]
    HEADER_SIZE = 13
    BUFFER_SIZE = 1024

    def __init__(
            self,
            server: str,
            port: int = 10051,
            timeout: Optional[float] = None,
            tls_connect: Optional[str] = None,
            tls_psk_identity: Optional[str] = None,
            tls_psk: Optional[str] = None,
            tls_psk_file: Optional[str] = None,
            tls_server_name: Optional[str] = None) -> None:
        self.server = server
        self.port = port
        self.timeout = timeout
        self.tls_connect = tls_connect
        self.tls_psk_identity = tls_psk_identity
        self.tls_psk = tls_psk
        self.tls_psk_file = tls_psk_file
        self.tls_server_name = tls_server_name

        self._validate_tls_options()

    def _validate_tls_options(self) -> None:
        if self.tls_connect is None:
            return

        if self.tls_connect != 'psk':
            raise ValueError("only tls_connect='psk' is supported")

        if not self.tls_psk_identity:
            raise ValueError('tls_psk_identity is required for PSK')

        if bool(self.tls_psk) == bool(self.tls_psk_file):
            raise ValueError(
                'set exactly one of tls_psk or tls_psk_file for PSK')

    def _create_payload(self, items: Sequence[ItemData]) -> bytes:
        payload = json.dumps({
            "request": "sender data",
            "data": items
        }).encode('utf-8')

        return payload

    def _create_packet(self, items: Sequence[ItemData]) -> bytes:
        # https://www.zabbix.com/documentation/current/en/manual/appendix/items/trapper
        payload = self._create_payload(items)
        packet = b'ZBXD' + b'\x01' + struct.pack('<II', len(payload), 0) + payload

        logging.debug(f'packet created, length: {len(packet)} bytes')
        return packet

    def _parse_response_header(self, header: bytes) -> Tuple[int, bool]:
        if len(header) != self.HEADER_SIZE:
            raise ValueError('invalid zabbix header size')

        if header[0:4] != b'ZBXD':
            raise ValueError('zabbix header not found or incorrect')

        flags = header[4]
        if not flags & 0x01:
            raise ValueError('invalid zabbix protocol version flag')

        # Zabbix header in 7.0 still uses 13 bytes:
        # magic(4) + flags(1) + datalen(4) + reserved(4)
        # For normal packets (flag 0x04 not set), reserved is 0.
        # For large packets (flag 0x04 set), datalen is 64-bit using both fields.
        data_len_low, data_len_high = struct.unpack('<II', header[5:])
        response_size = data_len_low + (data_len_high << 32)
        compressed = bool(flags & 0x02)

        logging.debug('response header received, response size %d bytes', response_size)
        if compressed:
            logging.debug('response payload is compressed')

        return response_size, compressed

    async def _write_data(self, writer: asyncio.StreamWriter, packet: bytes):
        writer.write(packet)
        await writer.drain()

    def _parse_response(self, response: bytes) -> ZabbixResponse:
        obj = json.loads(response)

        parsed_data = ZABBIX_RETURN_REGEX.match(obj['info'])

        response_object = ZabbixResponse(
            response=obj['response'],
            processed=int(parsed_data.group(1)),
            failed=int(parsed_data.group(2)),
            total=int(parsed_data.group(3)),
            seconds_spent=float(parsed_data.group(4))
        )
        logging.debug(f'parsed response: {response_object}')
        return response_object

    async def _read_response(self, reader: asyncio.StreamReader):
        header = await reader.readexactly(self.HEADER_SIZE)
        resp_size, compressed = self._parse_response_header(header)

        data = b''
        while len(data) < resp_size:
            buffer = await reader.read(min(self.BUFFER_SIZE, resp_size - len(data)))
            if not buffer:
                raise ConnectionError('connection closed before full response was received')
            data += buffer

        if compressed:
            data = zlib.decompress(data)

        resp = self._parse_response(data)
        return resp

    def _parse_psk(self) -> bytes:
        if self.tls_psk_file:
            raw = Path(self.tls_psk_file).read_text(encoding='utf-8').strip()
        else:
            raw = str(self.tls_psk).strip()

        if not raw:
            raise ValueError('PSK cannot be empty')

        try:
            return bytes.fromhex(raw)
        except ValueError as exc:
            raise ValueError('PSK must be a valid hex string') from exc

    def _create_psk_ssl_context(self) -> ssl.SSLContext:
        psk = self._parse_psk()
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        context.set_ciphers('PSK')

        if hasattr(context, 'set_psk_client_callback'):
            # Python 3.13+ OpenSSL callback API.
            context.set_psk_client_callback(lambda _hint: (self.tls_psk_identity, psk))
            return context

        try:
            from sslpsk3 import SSLPSKContext  # type: ignore[import-not-found]
        except Exception as exc:
            raise RuntimeError(
                "TLS-PSK requires Python 3.13+ or package 'sslpsk3'"
            ) from exc

        compat_context = SSLPSKContext(ssl.PROTOCOL_TLS_CLIENT)
        compat_context.check_hostname = False
        compat_context.verify_mode = ssl.CERT_NONE
        compat_context.set_ciphers('PSK')
        compat_context.psk = psk
        compat_context.psk_identity = self.tls_psk_identity.encode('utf-8')
        compat_context.set_psk_client_callback(lambda hint: (self.tls_psk_identity, psk))
        compat_context.maximum_version = ssl.TLSVersion.TLSv1_2
        return compat_context

    def _get_connection_kwargs(self) -> dict:
        kwargs = {}

        if self.timeout is not None:
            kwargs['ssl_handshake_timeout'] = self.timeout

        if self.tls_connect == 'psk':
            kwargs['ssl'] = self._create_psk_ssl_context()
            kwargs['server_hostname'] = self.tls_server_name or self.server

        return kwargs

    async def send(self, items=None) -> ZabbixResponse:
        # force items to an array
        if isinstance(items, ItemData):
            items = [items]

        if items is None:
            items = []

        logging.debug(
            f'sending {len(items)} metrics to {self.server}:{self.port}')

        packet = self._create_packet(items)

        connection = asyncio.open_connection(
            self.server,
            self.port,
            **self._get_connection_kwargs())

        reader, writer = await asyncio.wait_for(connection, timeout=self.timeout)

        await self._write_data(writer, packet)

        response = await self._read_response(reader)

        writer.close()
        await writer.wait_closed()

        return response



import asyncio
import json
import os
import random
import string
import time
import urllib.request

from dotenv import load_dotenv
import pytest

from zabbixasync.sender import AsyncSender, ItemData


class TestSend:
    """
    Integration tests for Zabbix sender against a running Zabbix 7 instance.
    First, it creates a host using the Zabbix API, then send various metrics.

    Required environment variables:
    - ZABBIX_HOST
    - ZABBIX_API_PORT
    - ZABBIX_SENDER_PORT
    - ZABBIX_USER
    - ZABBIX_PASS

    Optional TLS-PSK variables:
    - ZABBIX_TLS_CONNECT=psk
    - ZABBIX_TLS_PSK_IDENTITY
    - ZABBIX_TLS_PSK or ZABBIX_TLS_PSK_FILE
    """

    REQUIRED_ENV_VARS = [
        "ZABBIX_HOST",
        "ZABBIX_API_PORT",
        "ZABBIX_SENDER_PORT",
        "ZABBIX_USER",
        "ZABBIX_PASS",
    ]
    ZBX_UNABLE_CONFIG = "Unable to select configuration."

    def get_metrics_description(self):
        # type codes:
        # https://www.zabbix.com/documentation/current/en/manual/api/reference/item/object
        return [
            ("test.metric.text", 4, self.generate_random_string(255)),
            ("test.metric.unsigned", 3, int(random.random() * 10000)),
            ("test.metric.float", 0, random.random()),
        ]

    def generate_random_string(self, size: int) -> str:
        return "".join(random.choices(string.ascii_lowercase, k=size))

    def get_zabbix_metrics(self, number_of_metrics=1):
        metrics = []
        for i in range(number_of_metrics):
            hostname = self.get_test_hostname()
            desc = self.get_metrics_description()[i % 3]
            metrics.append(ItemData(host=hostname, key=desc[0], value=desc[2]))
        return metrics

    def get_test_hostname(self):
        return "async-sender-test-host"

    def get_sender(self):
        kwargs = {}
        tls_connect = os.getenv("ZABBIX_TLS_CONNECT")
        if tls_connect:
            kwargs["tls_connect"] = tls_connect
            kwargs["tls_psk_identity"] = os.getenv("ZABBIX_TLS_PSK_IDENTITY")
            tls_psk = os.getenv("ZABBIX_TLS_PSK")
            tls_psk_file = os.getenv("ZABBIX_TLS_PSK_FILE")
            if tls_psk:
                kwargs["tls_psk"] = tls_psk
            if tls_psk_file:
                kwargs["tls_psk_file"] = tls_psk_file

        return AsyncSender(
            os.environ["ZABBIX_HOST"],
            int(os.environ["ZABBIX_SENDER_PORT"]),
            **kwargs,
        )

    def get_api_url(self):
        return "".join(
            [
                "http://",
                os.environ["ZABBIX_HOST"],
                ":",
                os.environ["ZABBIX_API_PORT"],
                "/api_jsonrpc.php",
            ]
        )

    def add_default_headers(self, req: urllib.request.Request):
        req.add_header("Content-Type", "application/json-rpc")
        req.add_header("User-Agent", "zabbix-sender-async")

    def get_generic_request(self) -> dict:
        return {"jsonrpc": "2.0", "id": 1}

    def do_request(self, request_obj: dict) -> dict:
        api_url = self.get_api_url()
        data = json.dumps(request_obj).encode("utf-8")
        req = urllib.request.Request(api_url, data)
        self.add_default_headers(req)
        ret = urllib.request.urlopen(req, timeout=20)
        ret_str = ret.read().decode("utf-8")
        ret_json = json.loads(ret_str)

        if (
            "error" in ret_json
            and ret_json["error"]["code"] == 1
            and ret_json["error"]["message"] == self.ZBX_UNABLE_CONFIG
        ):
            # server may still be warming up
            time.sleep(10)
            return self.do_request(request_obj)
        if "error" in ret_json:
            raise Exception(ret_json["error"]["message"])

        return ret_json

    def do_login(self, user: str, password: str) -> str:
        request_json = self.get_generic_request()
        request_json["method"] = "user.login"
        request_json["params"] = {
            "username": user,
            "password": password,
        }

        ret_json = self.do_request(request_json)
        if "error" in ret_json:
            raise Exception(ret_json["error"])
        return ret_json["result"]

    def wait_api_ready(self) -> None:
        # Zabbix API version call does not require auth.
        request_json = self.get_generic_request()
        request_json["method"] = "apiinfo.version"
        request_json["params"] = {}
        last_error = None
        for _ in range(24):
            try:
                self.do_request(request_json)
                return
            except Exception as exc:
                last_error = exc
                time.sleep(5)
        raise RuntimeError(f"Zabbix API did not become ready: {last_error}")

    def get_groupid(self, auth: str) -> int:
        request_json = self.get_generic_request()
        request_json["method"] = "hostgroup.get"
        request_json["auth"] = auth
        request_json["params"] = {"output": ["groupid"], "limit": 1}
        ret_json = self.do_request(request_json)

        groups = ret_json["result"]
        if len(groups) == 0:
            raise Exception("cannot find any group to create the test host")
        return int(groups[0]["groupid"])

    def create_host(self, auth: str):
        hostname = self.get_test_hostname()
        groupid = self.get_groupid(auth)

        request_json = self.get_generic_request()
        request_json["auth"] = auth
        request_json["method"] = "host.create"
        request_json["params"] = {
            "host": hostname,
            "interfaces": [
                {
                    "type": 1,
                    "main": 1,
                    "useip": 1,
                    "ip": "127.0.0.1",
                    "dns": "",
                    "port": "10050"
                }
            ],
            "groups": [{"groupid": groupid}],
            "tls_accept": 3,
            "tls_connect": 2,
            "tls_psk_identity": "psk001",
            "tls_psk": "578ad7af47cdfc9f73b41e6ee4d68587b351506b70613b0c21758178bef20587"
        }

        ret_json = self.do_request(request_json)
        hostid = int(ret_json["result"]["hostids"][0])
        self.create_items(auth, hostid)

    def create_items(self, auth: str, host: int):
        for item in self.get_metrics_description():
            request_json = self.get_generic_request()
            request_json["auth"] = auth
            request_json["method"] = "item.create"
            request_json["params"] = {
                "name": item[0],
                "key_": item[0],
                "hostid": host,
                "type": 2,  # zabbix trapper
                "value_type": item[1],
            }
            self.do_request(request_json)

    def host_exists(self, auth: str):
        hostname = self.get_test_hostname()
        request_json = self.get_generic_request()
        request_json["auth"] = auth
        request_json["method"] = "host.get"
        request_json["params"] = {
            "filter": {"host": [hostname]},
            "output": ["hostid"],
            "limit": 1,
        }

        ret_json = self.do_request(request_json)
        return len(ret_json["result"]) > 0

    def validate_environment_variables(self):
        missing = [e for e in self.REQUIRED_ENV_VARS if e not in os.environ]
        if missing:
            pytest.skip(
                f"missing integration env vars: {missing}. "
                "Set them to run Zabbix integration tests."
            )

    def setup_host(self):
        self.wait_api_ready()
        auth = self.do_login(os.environ["ZABBIX_USER"], os.environ["ZABBIX_PASS"])
        if not self.host_exists(auth):
            self.create_host(auth)
            # wait for zabbix to persist host and items
            time.sleep(10)

    @pytest.fixture(autouse=True)
    def setup(self):
        load_dotenv()
        self.validate_environment_variables()
        self.setup_host()

    async def test_simple_send(self):
        sender = self.get_sender()
        metrics = self.get_zabbix_metrics()

        result = await sender.send(metrics)
        assert result is not None
        assert result.response == "success"
        assert result.total == 1
        assert result.processed == 1

    async def test_fail_send(self):
        sender = self.get_sender()
        invalid_data = ItemData("invalid_host", "invalid.metric", 0)
        result = await sender.send(invalid_data)

        assert result is not None
        assert result.response == "success"
        assert result.failed == 1

    async def test_big_chunk(self):
        metrics_count = 1000
        sender = self.get_sender()
        metrics = []

        dummy_metrics = self.get_zabbix_metrics(metrics_count)
        [metrics.append(dummy_metric) for dummy_metric in dummy_metrics]

        assert len(metrics) == metrics_count

        start = time.time()
        response = await asyncio.gather(sender.send(metrics))
        end = time.time()
        duration_first_send = end - start

        assert response[0] is not None
        assert response[0].response == "success"
        assert response[0].processed == metrics_count
        assert duration_first_send > 0

        tasks = [sender.send(metrics), sender.send(metrics), sender.send(metrics)]

        start = time.time()
        response = await asyncio.gather(*tasks)
        end = time.time()
        duration_second_send = end - start

        assert duration_second_send < (duration_first_send * 3)

    async def test_long_running_task(self):
        sleep_time = 3
        sender = self.get_sender()

        tasks = []
        tasks.append(asyncio.sleep(sleep_time))
        tasks.append(asyncio.sleep(sleep_time / 2))
        tasks.append(asyncio.sleep(sleep_time / 3))
        tasks.append(asyncio.sleep(sleep_time / 4))
        tasks.append(asyncio.sleep(sleep_time / 5))
        tasks.append(sender.send(self.get_zabbix_metrics()))

        start = time.time()
        await asyncio.gather(*tasks)
        end = time.time()

        execution_time = end - start
        assert execution_time <= sleep_time + 0.05

    async def test_big_metric(self):
        sender = self.get_sender()
        items = []

        for _ in range(5000):
            for metric in self.get_zabbix_metrics():
                items.append(metric)

        response = await sender.send(items)
        assert response is not None
        assert response.processed == 5000
        assert response.response == "success"

    async def test_psk_host(self):
        os.environ['ZABBIX_TLS_CONNECT'] = 'psk'
        os.environ['ZABBIX_TLS_PSK_IDENTITY'] = 'psk001'
        os.environ['ZABBIX_TLS_PSK'] = '578ad7af47cdfc9f73b41e6ee4d68587b351506b70613b0c21758178bef20587'

        sender = self.get_sender()
        metrics = self.get_zabbix_metrics()

        result = await sender.send(metrics)
        assert result is not None
        assert result.response == "success"
        assert result.total == 1
        assert result.processed == 1



import os
from tempfile import NamedTemporaryFile
from unittest import TestCase

from zabbixasync.sender import AsyncSender


class TestTlsPsk(TestCase):
    def test_validate_requires_identity(self):
        with self.assertRaises(ValueError):
            AsyncSender(
                "server",
                tls_connect="psk",
                tls_psk="1A2B",
            )

    def test_validate_requires_single_psk_source(self):
        with self.assertRaises(ValueError):
            AsyncSender(
                "server",
                tls_connect="psk",
                tls_psk_identity="id",
            )

        with self.assertRaises(ValueError):
            AsyncSender(
                "server",
                tls_connect="psk",
                tls_psk_identity="id",
                tls_psk="1A2B",
                tls_psk_file="dummy",
            )

    def test_parse_psk_from_string(self):
        sender = AsyncSender(
            "server",
            tls_connect="psk",
            tls_psk_identity="id",
            tls_psk="00112233445566778899AABBCCDDEEFF",
        )

        self.assertEqual(
            sender._parse_psk(),
            bytes.fromhex("00112233445566778899AABBCCDDEEFF"),
        )

    def test_parse_psk_from_file(self):
        with NamedTemporaryFile(
                mode="w+",
                encoding="utf-8",
                delete=False) as temp:
            temp.write("00112233445566778899AABBCCDDEEFF\n")
            temp.flush()
            temp_name = temp.name

        try:
            sender = AsyncSender(
                "server",
                tls_connect="psk",
                tls_psk_identity="id",
                tls_psk_file=temp_name,
            )

            self.assertEqual(
                sender._parse_psk(),
                bytes.fromhex("00112233445566778899AABBCCDDEEFF"),
            )
        finally:
            os.unlink(temp_name)

    def test_parse_psk_invalid_hex(self):
        sender = AsyncSender(
            "server",
            tls_connect="psk",
            tls_psk_identity="id",
            tls_psk="zzzz",
        )

        with self.assertRaises(ValueError):
            sender._parse_psk()



import json
from unittest import TestCase

from zabbixasync.sender import ItemData


class TestItemData(TestCase):
    def test_init(self):
        expected = '{"host": "localhost", "key": "test", "value": 123456}'
        id = ItemData('localhost', 'test', 123456)
        dump = json.dumps(id)
        self.assertEqual(dump, expected)

    def test_init_with_ns_clock(self):
        expected = '{"host": "localhost", "key": "test", "value": 123456, \
"clock": 123456789, "ns": 123456789}'

        id = ItemData('localhost', 'test', 123456, 123456789, 123456789)
        dump = json.dumps(id)
        self.assertEqual(dump, expected)



from unittest import TestCase

from zabbixasync.sender import AsyncSender


class TestResponseHeader(TestCase):
    def setUp(self):
        self.sender = AsyncSender('server')

    def test_packet_creation(self):
        header_received = b'ZBXD\x01[\x00\x00\x00\x00\x00\x00\x00'
        expected = 91
        res, compressed = self.sender._parse_response_header(header_received)

        self.assertEqual(res, expected)
        self.assertFalse(compressed)

    def test_packet_creation_compressed(self):
        header_received = b'ZBXD\x03[\x00\x00\x00\x00\x00\x00\x00'
        expected = 91
        res, compressed = self.sender._parse_response_header(header_received)

        self.assertEqual(res, expected)
        self.assertTrue(compressed)

    def test_packet_header_incorrect(self):
        with self.assertRaises(ValueError):
            header_received_incorrect = \
                b'EEEE\x01[\x00\x00\x00\x00\x00\x00\x00'
            self.sender._parse_response_header(header_received_incorrect)

    def test_packet_header_slightly_incorrect(self):
        with self.assertRaises(ValueError):
            header_received_incorrect = \
                b'ZBXD\x00[\x00\x00\x00\x00\x00\x00\x00'
            self.sender._parse_response_header(header_received_incorrect)



from unittest import TestCase

from zabbixasync.sender import AsyncSender


class TestResponsePayload(TestCase):
    def setUp(self):
        self.sender = AsyncSender('server')

    def test_response_parse(self):
        payload_received = b'{"response":"success","info":"processed: 1; \
failed: 2; total: 3; seconds spent: 0.123456"}'
        response_sucess_expected = 'success'
        response_processed_expected = 1
        response_failed_expected = 2
        response_total_expected = 3
        response_seconds_spent_expected = 0.123456

        resp = self.sender._parse_response(payload_received)
        self.assertEqual(resp.response, response_sucess_expected)
        self.assertEqual(resp.processed, response_processed_expected)
        self.assertEqual(resp.failed, response_failed_expected)
        self.assertEqual(resp.total, response_total_expected)
        self.assertEqual(resp.seconds_spent, response_seconds_spent_expected)



from unittest import TestCase

from zabbixasync.sender import AsyncSender, ItemData


class TestPayload(TestCase):
    def test_payload_content(self):
        expected = b'{"request": "sender data", "data": \
[{"host": "localhost", "key": "test", "value": 123456}]}'

        data = ItemData('localhost', 'test', 123456)
        items = [data]
        sender = AsyncSender('server')
        payload = sender._create_payload(items)

        self.assertEqual(payload, expected)



import asyncio

from zabbixasync.sender import AsyncSender, ItemData

ZABBIX_SERVER = "127.0.0.1"  # update if needed
ZABBIX_PORT = 10051
ZABBIX_HOST = "async-sender-test-host"
PSK_IDENTITY = "psk001"
PSK_HEX = "578ad7af47cdfc9f73b41e6ee4d68587b351506b70613b0c21758178bef20587"
LONG_TEXT_VALUE = (
    "This is a long trapper text payload used to verify sender behavior with "
    "bigger string values in Zabbix. "
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod "
    "tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim "
    "veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea "
    "commodo consequat."
)


async def main() -> None:
    sender = AsyncSender(
        ZABBIX_SERVER,
        ZABBIX_PORT,
        tls_connect="psk",
        tls_psk_identity=PSK_IDENTITY,
        tls_psk=PSK_HEX,
    )
    items = [
        ItemData(
            host=ZABBIX_HOST,
            key="test.trapper.int",
            value=42,
        ),
        ItemData(
            host=ZABBIX_HOST,
            key="test.trapper.text",
            value=LONG_TEXT_VALUE,
        ),
    ]
    result = await sender.send(items)
    print(result)


if __name__ == "__main__":
    asyncio.run(main())
